from .version import version as __version__

from .comparisontools import (count_matching_events, compute_agreement_score, count_match_spikes,
        make_agreement_scores, make_possible_match, make_best_match, make_hungarian_match, 
        do_score_labels, compare_spike_trains, do_confusion_matrix, do_count_score, compute_performence)
from .symmetricsortingcomparison import compare_two_sorters, SymmetricSortingComparison
from .groundtruthcomparison import compare_sorter_to_ground_truth, GroundTruthComparison
from .multisortingcomparison import compare_multiple_sorters, MultiSortingComparison

from .collisioncomparison import CollisionGTComparison

from .groundtruthstudy import GroundTruthStudy
